/******************************************************************************

  Copyright(c) 2006 Intel Corporation. All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of version 2 of the GNU General Public License as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110,
  USA

  The full GNU General Public License is included in this distribution
  in the file called LICENSE.

  Contact Information:
  James P. Ketrenos <ipw2100-admin@linux.intel.com>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR
  97124-6497

******************************************************************************/

/******************************************************************************

  To build:

	% gcc -I/usr/src/linux/include -Wall -o sim_rx sim_rx.c

  To run:

	% ./sim_rx --help

******************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <math.h>
#include <ctype.h>

typedef signed char s8;
typedef unsigned char u8;
typedef unsigned short u16;
typedef signed short s16;
typedef unsigned int u32;
typedef signed int s32;
typedef unsigned long long u64;
typedef signed long long s64;

#include <net/ieee80211_radiotap.h>

/*****************************************************************************
 *
 * IEEE 802.11 Declarations
 *
 * We define these here as including net/ieee80211.h has too many other
 * typedef dependencies.  Also, depending on the net/ieee80211.h on the system,
 * some of these types may not be defined.
 *
 */
#define IEEE80211_MAC_LEN       6
#define IEEE80211_3ADDR_LEN     24

#define IEEE80211_FCTL_FTYPE	0x000c
#define IEEE80211_FCTL_STYPE	0x00f0
#define IEEE80211_FCTL_TODS	0x0100

#define IEEE80211_FTYPE_MGMT	0x0000

#define IEEE80211_STYPE_PROBE_RESP 0x0050
#define IEEE80211_STYPE_BEACON	   0x0080
#define IEEE80211_STYPE_DEAUTH     0x00C0

struct ieee80211_hdr_3addr {
	__le16 frame_ctl;
	__le16 duration_id;
	u8 addr1[IEEE80211_MAC_LEN];
	u8 addr2[IEEE80211_MAC_LEN];
	u8 addr3[IEEE80211_MAC_LEN];
	__le16 seq_ctl;
	u8 payload[0];
} __attribute__ ((packed));

/* IEEE 802.11h Declarations */
#define IEEE80211_STYPE_ACTION 0x00D0

enum {
	IEEE80211_IE_CSA = 37,
};

enum {
	IEEE80211_ACTION_SPECTRUM_MGMT = 0,
};

enum {
	IEEE80211_ACTION_CATEGORY_CHANNEL_SWITCH = 4,
};

struct ieee80211_channel_switch {
	u8 id;
	u8 len;
	u8 mode;
	u8 channel;
	u8 count;
} __attribute__ ((packed));

struct ieee80211h_csa {
	struct ieee80211_hdr_3addr header;
	u8 category;
	u8 action;
	struct ieee80211_channel_switch channel_switch;
} __attribute__ ((packed));

struct ieee80211_deauth {
	struct ieee80211_hdr_3addr header;
	__le16 reason;
} __attribute__ ((packed));

/*****************************************************************************/

#define VERSION "0.1"
#define COPYRIGHT "Copyright (C) 2006, Intel Corporation."

struct ipw_rt_hdr {
	struct ieee80211_radiotap_header rt_hdr;
	u64 rt_tsf;
	u8 rt_flags;
	u8 rt_rate;
	u16 rt_channel;
	u16 rt_chbitmask;
	s8 rt_dbmsignal;
	s8 rt_dbmnoise;
	u8 rt_antenna;
	u8 payload[0];
} __attribute__ ((packed));

/* 3945 Simulation Sysfs Interface declarations */

struct sim_frame {
	struct ipw_rt_hdr rt;
	union {
		struct ieee80211_deauth deauth;
		struct ieee80211h_csa csa;
	} type;
} __attribute__ ((packed));

#define SYSFS_DRIVER "/sys/bus/pci/drivers/ipw3945"

struct priv {
	char sysfs_path[PATH_MAX];
	u8 csa_channel;
	u8 channel;
	u8 src[IEEE80211_MAC_LEN];
	u8 dst[IEEE80211_MAC_LEN];
	u8 bss[IEEE80211_MAC_LEN];
	int set_channel;
	int set_bss;
	int set_dst;
	int set_src;
	int set_type;
	int set_csa_channel;
};

int find_sysfs(struct priv *priv)
{
	DIR *dir = NULL;
	char cwd[PATH_MAX];
	struct dirent *entry;
	int rc = 0;

	char *path = priv->sysfs_path;
	size_t len = sizeof(priv->sysfs_path);

	getcwd(cwd, sizeof(cwd));

	dir = opendir(SYSFS_DRIVER);
	if (dir == NULL) {
		rc = -errno;
		fprintf(stderr, "opening " SYSFS_DRIVER ":\n"
			"%s (%d)\n", strerror(errno), errno);
		goto exit;
	}

	if (chdir(SYSFS_DRIVER)) {
		rc = -errno;
		fprintf(stderr, "changing to " SYSFS_DRIVER ":\n"
			"%s (%d)\n", strerror(errno), errno);
		goto exit;
	}

	entry = readdir(dir);
	while (entry) {
		struct stat info;

		if (!strcmp(entry->d_name, ".") || !strcmp(entry->d_name, ".."))
			goto next_loop;

		if (stat(entry->d_name, &info))
			goto next_loop;

		if (S_ISDIR(info.st_mode)) {
			if (chdir(entry->d_name)) {
				rc = -errno;
				fprintf(stderr, "changing to %s:\n"
					"%s (%d)\n",
					entry->d_name, strerror(errno), errno);

				goto exit;
			}

			if (stat("rx", &info)) {
				chdir(SYSFS_DRIVER);
				goto next_loop;
			}

			break;
		}
	      next_loop:
		entry = readdir(dir);
	}

	if (entry) {
		snprintf(path, len, SYSFS_DRIVER "/%s", entry->d_name);
		rc = 0;
		goto exit;
	}

	rc = -ENOENT;
	goto exit;

      exit:

	if (dir)
		closedir(dir);

	chdir(cwd);

	return rc;
}

int usage(char *argv[], int ret)
{
	if (ret)
		fprintf(stderr, "Invalid parameter %d: %s.\n\n",
			ret, argv[ret]);

	fprintf(stdout, "usage: sim_rx [--OPTION]\n"
		"options:\n"
		"  --type=[N]\t\tSet frame type, where type is one of:\n"
		"    0: CSA\n"
		"    1: DEAUTH\n"
		"  --channel=channel\tSource channel for simulated Rx.\n"
		"  --csa-channel=channel\tTarget channel used in CSA.\n"
		"  --src=SRC_MAC\t\tSource MAC in form 00:00:00:00:00:00\n"
		"  --dst=DST_MAC\t\tDestination MAC in form 00:00:00:00:00:00\n"
		"  --bss=BSS_MAC\t\tNetwowrk MAC in form 00:00:00:00:00:00\n"
		"  --version\t\tDisplay version information\n"
		"  --help\t\tView this help message\n");

	return -ret;
}

enum {
	PARAM_NONE = 0,
	PARAM_VALUE = 1,
};

int tolower(int c)
{
	return (c >= 'A' && c <= 'Z') ? c - 'A' : c;
}

int tohex(int c)
{
	return (c >= '0' && c <= '9') ? c - '0' : c - 'a' + 0xA;
}

int parse_mac(const char *param, u8 * dst)
{
	int i;
	for (i = 0; i < IEEE80211_MAC_LEN; i++) {
		if (*param == '\0')
			break;
		dst[i] = tohex(tolower(*param++)) << 4;

		if (*param == '\0')
			break;
		dst[i] |= tohex(tolower(*param++));

		if (*param != '\0' && *param == ':')
			param++;
	}

	return (i != IEEE80211_MAC_LEN);
}

inline int param_check(const char *param, const char *value, int type)
{
	int len = strlen(value);
	int comp = strncmp(param, value, len);

	if (comp)
		return 0;

	if (type == PARAM_VALUE)	/* = */
		return (param[len] != '=') ? -1 : len + 1;

	/* no = */

	return (param[len] != '\0') ? -1 : len;
}

int version()
{
	fprintf(stdout, "sim_rx - Packet Rx simulation\n");
	fprintf(stdout, "%s\n", COPYRIGHT);
	fprintf(stdout, "version: " VERSION "\n");
	return 0;
}

int init_radiotap(struct priv *priv, struct ipw_rt_hdr *rt)
{
	if (!priv->set_channel) {
		fprintf(stderr, "--channel must be set.\n");
		return 0;
	}
	rt->rt_channel = priv->channel;
	return (sizeof(*rt));
}

void init_ieee80211_hdr_3addr(struct priv *priv,
			      struct ieee80211_hdr_3addr *hdr, __le16 frame_ctl)
{
	hdr->duration_id = 0;
	hdr->seq_ctl = 0;
	hdr->frame_ctl = frame_ctl;

	memcpy(hdr->addr3, priv->src, IEEE80211_MAC_LEN);
	memcpy(hdr->addr1, priv->dst, IEEE80211_MAC_LEN);
	memcpy(hdr->addr2, priv->bss, IEEE80211_MAC_LEN);
}

int init_csa(struct priv *priv, struct ieee80211h_csa *frame)
{
	if (!priv->set_bss || !priv->set_dst || !priv->set_src ||
	    !priv->set_channel) {
		fprintf(stderr, "--bss, --dst, --src, and "
			"--csa-channel must all be set.\n");
		return 0;
	}

	init_ieee80211_hdr_3addr(priv, &frame->header,
				 IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_ACTION);

	frame->category = IEEE80211_ACTION_SPECTRUM_MGMT;
	frame->action = IEEE80211_ACTION_CATEGORY_CHANNEL_SWITCH;

	frame->channel_switch.id = IEEE80211_IE_CSA;
	frame->channel_switch.len = 3;

	/* New Channel */
	frame->channel_switch.channel = 40;

	/* Mode:
	 * 1 - No more Tx allowed until on the new channel
	 * 0 - No Tx rerequiement on the receiving client
	 */
	frame->channel_switch.mode = 1;

	/* Channel Switch Count:
	 * >1 - Number of beacon transmission intervals until switch.
	 * 1  - Occur immediately before next TBTT.
	 * 0  - Occur immediately
	 */
	frame->channel_switch.count = 0;

	return sizeof(*frame);
}

int init_deauth(struct priv *priv, struct ieee80211_deauth *frame)
{
	if (!priv->set_bss || !priv->set_dst || !priv->set_src) {
		fprintf(stderr, "--bss, --dst, and --src "
			"must all be set.\n");
		return 0;
	}
	init_ieee80211_hdr_3addr(priv, &frame->header,
				 IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_DEAUTH);

	frame->reason = 5;

	return sizeof(*frame);
}

int send_frame(struct priv *priv, struct sim_frame *frame, int len)
{
	char full_path[PATH_MAX];
	int fd, rc = 0;

	snprintf(full_path, sizeof(full_path), "%s/rx", priv->sysfs_path);

	fprintf(stdout, "Sending %d bytes.\n", len);

	fd = open(full_path, O_WRONLY);
	if (fd < 0) {
		fprintf(stderr, "opening '%s':\n%s (%d)\n",
			full_path, strerror(errno), errno);
		return -errno;
	}

	rc = write(fd, frame, sizeof(frame->rt) + len);
	if (rc == -1) {
		fprintf(stderr, "write failed: %s (%d)\n", strerror(errno),
			errno);
		rc = -errno;
		close(fd);
		return rc;
	}

	close(fd);

	if (rc != sizeof(frame->rt) + len) {
		fprintf(stderr, "error writing %zd bytes (%d written)\n",
			sizeof(frame->rt) + len, rc);
		return -EINVAL;
	}

	return 0;
}

enum {
	CSA = 0,
	DEAUTH = 1,
};

int main(int argc, char *argv[])
{
	struct priv priv;
	int rc = 0, pos, i = 0;
	struct sim_frame frame;
	int len;
	int frame_type = DEAUTH;

	memset(&priv, 0, sizeof(priv));
	while (i < argc - 1) {
		i++;

		if (argv[i][0] != '-' || argv[i][1] != '-')
			return usage(argv, i);

		pos = param_check(&argv[i][2], "type", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			frame_type = strtol(&argv[i][2 + pos], NULL, 0);
			priv.set_type = 1;
			continue;
		}

		pos = param_check(&argv[i][2], "csa-channel", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			priv.csa_channel = strtol(&argv[i][2 + pos], NULL, 0);
			priv.set_csa_channel = 1;
			continue;
		}

		pos = param_check(&argv[i][2], "channel", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			priv.channel = strtol(&argv[i][2 + pos], NULL, 0);
			priv.set_channel = 1;
			continue;
		}

		pos = param_check(&argv[i][2], "src", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			if (parse_mac(&argv[i][2 + pos], priv.src))
				return usage(argv, i);
			priv.set_src = 1;
			continue;
		}

		pos = param_check(&argv[i][2], "dst", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			if (parse_mac(&argv[i][2 + pos], priv.dst))
				return usage(argv, i);
			priv.set_dst = 1;
			continue;
		}

		pos = param_check(&argv[i][2], "bss", PARAM_VALUE);
		if (pos < 0)
			return usage(argv, i);
		if (pos) {
			if (parse_mac(&argv[i][2 + pos], priv.bss))
				return usage(argv, i);
			priv.set_bss = 1;
			continue;
		}

		pos = param_check(&argv[i][2], "help", PARAM_NONE);
		if (pos < 0)
			return usage(argv, i);
		if (pos)
			return usage(argv, 0);

		pos = param_check(&argv[i][2], "version", PARAM_NONE);
		if (pos < 0)
			return usage(argv, i);
		if (pos)
			return version();

		return usage(argv, i);
	}

	rc = find_sysfs(&priv);
	if (rc) {
		fprintf(stderr, "Could not find " SYSFS_DRIVER "/*/tx\n");
		goto exit;
	}

	if (!init_radiotap(&priv, &frame.rt))
		return -1;

	if (!priv.set_type) {
		fprintf(stderr, "--type must be set.\n");
		return 0;
	}

	switch (frame_type) {
	case CSA:
		len = init_csa(&priv, &frame.type.csa);
		break;
	case DEAUTH:
		len = init_deauth(&priv, &frame.type.deauth);
		break;
	}

	if (len == 0)
		return -1;

	send_frame(&priv, &frame, len);

      exit:
	return rc;
}
